library flutter_bloc_architecture;

import 'dart:async';

import 'package:flutter/widgets.dart';

class BlocInstance {
  static BlocStorageInterface _storage;

  static BlocStorage get storage {
    assert(BlocInstance._storage != null, "Bloc storage is not initialized. To do this, call: Bloc.initialize() in function main()");
    return BlocInstance._storage;
  }

  static void initialize(BlocStorageInterface storage) {
    if (BlocInstance._storage == null) {
      assert(storage != null);
      BlocInstance._storage = storage;
    }
  }

  static void deinitialize() {
    if (_storage != null) {
      _storage.dispose();
      _storage = null;
    }
  }
}

abstract class BlocState {
  void dispose() {}
}

abstract class BlocEvent {}

abstract class BlocAction<STATE extends BlocState> {
  STATE action(STATE state, Bloc<STATE> bloc);
}

typedef void BlocOnStateChanged<STATE extends BlocState>(BlocState state, Bloc<STATE> bloc);
typedef void BlocOnEvent<STATE extends BlocState>(BlocEvent event, Bloc<STATE> bloc);
typedef void BlocOnError<STATE extends BlocState>(Object error, Bloc<STATE> bloc);
typedef void BlocOnUnsubscribe<STATE extends BlocState>(Bloc<STATE> bloc);

typedef Widget BlocOnWidgetInitState<STATE extends BlocState>(STATE state, Bloc<STATE> bloc);
typedef Widget BlocOnWidgetDidChangeDependencies<STATE extends BlocState>(BuildContext context, STATE state, Bloc<STATE> bloc);
typedef Widget BlocOnWidgetBuild<STATE extends BlocState>(BuildContext context, STATE state, Bloc<STATE> bloc);
typedef Widget BlocOnWidgetDispose<STATE extends BlocState>(BuildContext context, STATE state, Bloc<STATE> bloc);
typedef void BlocOnWidgetEvent<STATE extends BlocState>(BuildContext context, BlocEvent event, Bloc<STATE> bloc);
typedef void BlocOnWidgetError<STATE extends BlocState>(BuildContext context, Object error, Bloc<STATE> bloc);
typedef void BlocOnWidgetUnsubscribe<STATE extends BlocState>(BuildContext context, Bloc<STATE> bloc);

class Bloc<STATE extends BlocState> {
  STATE _state;
  StreamController<BlocAction> _privateStreamController;
  StreamController<Object> _broadcastStreamController;

  Bloc(STATE initialState) {
    _privateStreamController = StreamController<BlocAction>.broadcast(sync: false);
    _privateStreamController.stream.listen(_onAction);

    _broadcastStreamController = StreamController<Object>.broadcast(sync: false);

    this._state = initialState;
  }

  void _onAction(BlocAction action) {
    final state = action.action(this._state, this);
    if (state != null) {
      this._state = state;
      _broadcastStreamController.sink.add(this._state);
    }
  }

  void enqueueAction(BlocAction action) {
    if (action != null) {
      _privateStreamController.sink.add(action);
    }
  }

  void enqueueEvent(BlocEvent event) {
    if (event != null) {
      _broadcastStreamController.sink.add(event);
    }
  }

  void enqueueError(Object error) {
    if (error != null) {
      _broadcastStreamController.sink.addError(error);
    }
  }

  void dispose() {
    if (_privateStreamController != null) {
      _privateStreamController.close();
    }
    if (_broadcastStreamController != null) {
      _broadcastStreamController.close();
    }
    if (_state != null) {
      _state.dispose();
    }
  }

  STATE get state => _state;
}

abstract class BlocStorageInterface {
  void dispose();
}

class BlocStorage implements BlocStorageInterface {
  final List<Bloc<BlocState>> _list;

  BlocStorage() : _list = List();

  void dispose() {
    for (final bloc in _list) {
      bloc.dispose();
    }
    _list.clear();
  }

  Bloc<STATE> create<STATE extends BlocState>(STATE state) {
    final Bloc<STATE> bloc = Bloc<STATE>(state);
    _list.add(bloc);
    return bloc;
  }

  void destroy(Bloc<BlocState> bloc) {
    if (bloc != null) {
      _list.remove(bloc);
      bloc.dispose();
    }
  }

  Bloc<STATE> get<STATE extends BlocState>(Type type) {
    return _list.firstWhere((bloc) {
      return (bloc._state.runtimeType == type);
    }, orElse: () {
      return null;
    });
  }

  Bloc<STATE> getOrCreate<STATE extends BlocState>(STATE state) {
    return _list.firstWhere((bloc) {
      return (bloc is Bloc<STATE>);
    }, orElse: () {
      final bloc = create(state);
      return bloc;
    });
  }

  static BlocStorage get storage {
    final storage = BlocInstance.storage;
    assert(storage is BlocStorage);
    return storage;
  }

  static void initialize() {
    BlocInstance.initialize(BlocStorage());
  }

  static void deinitialize() {
    BlocInstance.deinitialize();
  }
}

class BlocObserver<STATE extends BlocState> {
  Bloc<STATE> _bloc;
  StreamSubscription<Object> _streamSubscription;

  BlocObserver<STATE> subscribe({
    Bloc<STATE> bloc,
    BlocOnStateChanged onStateChanged,
    BlocOnEvent onEvent,
    BlocOnError onError,
    BlocOnUnsubscribe onUnsubscribe,
  }) {
    unsubscribe();

    this._bloc = bloc;
    if (this._bloc == null) {
      return this;
    }

    _streamSubscription = _bloc._broadcastStreamController.stream.listen(
      (data) {
        if (data is BlocState) {
          if (onStateChanged != null) {
            onStateChanged(data, _bloc);
          }
        } else if (data is BlocEvent) {
          if (onEvent != null) {
            onEvent(data, _bloc);
          }
        }
      },
      onError: (error) {
        if (onError != null) {
          onError(error, _bloc);
        }
      },
      onDone: () {
        if (onUnsubscribe != null) {
          onUnsubscribe(_bloc);
        }
      },
      cancelOnError: false,
    );

    return this;
  }

  BlocObserver<STATE> unsubscribe() {
    if (_streamSubscription != null) {
      _streamSubscription.cancel();
    }
    return this;
  }
}

class BlocWidgetObserver<STATE extends BlocState> extends StatefulWidget {
  final Bloc<STATE> bloc;
  final BlocOnWidgetInitState<STATE> onInitState;
  final BlocOnWidgetDidChangeDependencies<STATE> onDidChangeDependencies;
  final BlocOnWidgetBuild<STATE> onBuild;
  final BlocOnWidgetDispose<STATE> onDispose;
  final BlocOnWidgetEvent<STATE> onEvent;
  final BlocOnWidgetError<STATE> onError;
  final BlocOnWidgetUnsubscribe<STATE> onUnsubscribe;

  BlocWidgetObserver({
    @required this.bloc,
    this.onInitState,
    this.onDidChangeDependencies,
    @required this.onBuild,
    this.onDispose,
    this.onEvent,
    this.onError,
    this.onUnsubscribe,
  });

  @override
  State<StatefulWidget> createState() {
    return _BlocWidgetObserverState<STATE>();
  }
}

class _BlocWidgetObserverState<STATE extends BlocState> extends State<BlocWidgetObserver> {
  final BlocObserver<STATE> observer;

  _BlocWidgetObserverState() : observer = BlocObserver<STATE>();

  @override
  void initState() {
    if(widget.onInitState != null) {
      widget.onInitState(widget.bloc.state, widget.bloc);
    }
    observer.subscribe(
      bloc: widget.bloc,
      onStateChanged: _onStateChanged,
      onEvent: _onEvent,
      onError: _onError,
      onUnsubscribe: _onUnsubscribe,
    );
    super.initState();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    if(widget.onDidChangeDependencies != null) {
      widget.onDidChangeDependencies(context, widget.bloc.state, widget.bloc);
    }
  }

  @override
  Widget build(BuildContext context) {
    return widget.onBuild(context, observer._bloc._state, observer._bloc);
  }

  @override
  void dispose() {
    observer.unsubscribe();
    if(widget.onDispose != null) {
      widget.onDispose(context, widget.bloc.state, widget.bloc);
    }
    super.dispose();
  }

  void _onStateChanged(BlocState state, Bloc<STATE> bloc) {
    setState(() {});
  }

  void _onEvent(BlocEvent event, Bloc<STATE> bloc) {
    if (widget.onEvent != null) {
      widget.onEvent(context, event, bloc);
    }
  }

  void _onError(Object error, Bloc<STATE> bloc) {
    if (widget.onError != null) {
      widget.onError(context, error, bloc);
    }
  }

  void _onUnsubscribe(Bloc<STATE> bloc) {
    if (widget.onUnsubscribe != null) {
      widget.onUnsubscribe(context, bloc);
    }
  }
}
